
<a href="{{ pathto(master_doc) }}">
    <img class="logo" src="{{ pathto('_static/zeep-logo.png', 1) }}">
</a>

<p>Zeep is a modern SOAP client for Python</p>
<p>
  <iframe src="http://ghbtns.com/github-btn.html?user=mvantellingen&repo=python-zeep&type=watch&count=true&size=large"
    allowtransparency="true" frameborder="0" scrolling="0" width="200px" height="35px"></iframe>
</p>

<h3>Links</h3>
<ul>
  <li><a href="http://github.com/mvantellingen/python-zeep">Zeep @ GitHub</a></li>
  <li><a href="http://pypi.python.org/pypi/zeep">Zeep @ PyPI</a></li>
  <li><a href="http://github.com/mvantellingen/python-zeep/issues">Issue Tracker</a></li>
</ul>

